
/*
 * Glue library for DAHDI ioctl calls
 *
 * Written by Suporte <suporte@aligera.com.br>
 *
 * Copyright (C) 2012 Aligera Equipamentos Digitais LTDA.
 *
 * All rights reserved.
 *
 * Uses code from dahdi_tool, don't remove the GPL disclaimer below.
 */

/*
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2 as published by the
 * Free Software Foundation. See the LICENSE file included with
 * this program for more details.
 */

#include <Python.h>

#include <sys/ioctl.h>
#include <unistd.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <dahdi/user.h>
#include <dahdi/ap_user.h>

int ctl;

/**
 * For some obscure reason, it's kind of impossible to know
 * what is the channel number of a span's channel. So we do this...
 */
static int find_first_channel(int span)
{
    int res, i, channels;
    struct dahdi_spaninfo s;

    if( span >= DAHDI_MAX_SPANS )
        return -1;

    for(i = 1, channels = 0; i < span; i++) {
        s.spanno = i;
        res = ioctl(ctl, DAHDI_SPANSTAT, &s);
        if (res) return -1;

        channels += s.totalchans;
    }
    return channels+1;
}

static int dahdi_indirect_ioctl(int span, int cmd, void* data)
{
    struct dahdi_indirect_data ind;

    ind.chan = span;
    ind.op = cmd;
    ind.data = data;

    return ioctl(ctl, DAHDI_INDIRECT, &ind);
}

static PyObject *
ap400_find_first_channel(PyObject *self, PyObject *args)
{
    int span, chan;

    if (!PyArg_ParseTuple(args, "i", &span))
        return NULL;

    chan = find_first_channel(span);
    if( chan < 0 )
        return NULL;
    return Py_BuildValue("i", chan);
}

static char *_ap400_alarm_string(int alarm)
{
    static char alarms[50];
    strcpy(alarms, "");
    if ((alarm&(AP4_ALARM_LOS|AP4_ALARM_AIS|AP4_ALARM_BFAE|AP4_ALARM_RAI|AP4_ALARM_MFAE|AP4_ALARM_CAS))) {
        if (alarm & AP4_ALARM_LOS)
            strcat(alarms,"LOS/");
        if (alarm & AP4_ALARM_AIS)
            strcat(alarms, "AIS/");
        if (alarm & AP4_ALARM_BFAE)
            strcat(alarms, "BFAE/");
        if (alarm & AP4_ALARM_RAI)
            strcat(alarms,"RAI/");
        if (alarm & AP4_ALARM_MFAE)
            strcat(alarms,"MFAE/");
        if (alarm & AP4_ALARM_CAS)
            strcat(alarms, "CAS/");
        if (!strlen(alarms))
            strcat(alarms, "UUU/");
        if (strlen(alarms)) {
            /* Strip trailing / */
            alarms[strlen(alarms)-1]='\0';
        }
    } else {
        strcpy(alarms, "OK");
    }
    return alarms;
}

static PyObject *
ap400_alarms(PyObject *self, PyObject *args)
{
    int alarms = 0, res;
    int span, chan;
    PyObject *result, *member;

    if (!PyArg_ParseTuple(args, "i", &span))
        return NULL;

    chan = find_first_channel(span);
    res = dahdi_indirect_ioctl(chan, AP4_GET_ALARMS, &alarms);
    if( res ) return NULL;

    result = PyList_New (0);
    if (result == NULL)
        return NULL;

    member = Py_BuildValue("i", alarms);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyList_Append (result, member);
    Py_XDECREF (member);

    member = Py_BuildValue("s", _ap400_alarm_string(alarms));
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyList_Append (result, member);
    Py_XDECREF (member);

    return result;
}

static PyObject *
ap400_hwinfo(PyObject *self, PyObject *args)
{
    int res;
    int span, chan;
    struct ap_ioctl_hwinfo hw_info;
    PyObject *result, *member;

    if (!PyArg_ParseTuple(args, "i", &span))
        return NULL;

    memset(&hw_info, '\0', sizeof(hw_info));
    chan = find_first_channel(span);
    res = dahdi_indirect_ioctl(chan, AP_GET_HWINFO, &hw_info);
    if( res ) return NULL;

    result = PyDict_New ();
    if (result == NULL)
        return NULL;

    member = Py_BuildValue("I", hw_info.irq);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "irq", member);
    Py_XDECREF (member);

    member = Py_BuildValue("I", hw_info.fpga_major);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "fpga_major", member);
    Py_XDECREF (member);

    member = Py_BuildValue("I", hw_info.fpga_minor);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "fpga_minor", member);
    Py_XDECREF (member);

    return result;
}

#define A400STATS_ADDMEMBER(dic, _struct, _member) { \
    member = Py_BuildValue("I", _struct._member); \
    if( member == NULL ) { \
        Py_XDECREF (dic); \
        return NULL; \
    } \
    PyDict_SetItemString (dic, #_member, member); \
    Py_XDECREF (member); \
}

static PyObject *
ap400_spats(PyObject *self, PyObject *args)
{
    int res;
    int span, chan;
    struct ap4_ioctl_span_stats span_stats;
    PyObject *result, *member;

    if (!PyArg_ParseTuple(args, "i", &span))
        return NULL;

    memset(&span_stats, '\0', sizeof(span_stats));
    chan = find_first_channel(span);
    res = dahdi_indirect_ioctl(chan, AP4_GET_STATS, &span_stats);
    if( res ) return NULL;

    result = PyDict_New ();
    if (result == NULL)
        return NULL;

    A400STATS_ADDMEMBER(result, span_stats, span);
    A400STATS_ADDMEMBER(result, span_stats, seconds);
    A400STATS_ADDMEMBER(result, span_stats, slipcount);
    A400STATS_ADDMEMBER(result, span_stats, irqmisses);
    A400STATS_ADDMEMBER(result, span_stats, los_count);
    A400STATS_ADDMEMBER(result, span_stats, los_seconds);
    A400STATS_ADDMEMBER(result, span_stats, ais_count);
    A400STATS_ADDMEMBER(result, span_stats, ais_seconds);
    A400STATS_ADDMEMBER(result, span_stats, rai_count);
    A400STATS_ADDMEMBER(result, span_stats, rai_seconds);
    A400STATS_ADDMEMBER(result, span_stats, bfae_count);
    A400STATS_ADDMEMBER(result, span_stats, bfae_seconds);
    A400STATS_ADDMEMBER(result, span_stats, mfae_count);
    A400STATS_ADDMEMBER(result, span_stats, mfae_seconds);
    A400STATS_ADDMEMBER(result, span_stats, crc_count);
    A400STATS_ADDMEMBER(result, span_stats, bpv_count);
    A400STATS_ADDMEMBER(result, span_stats, cas_count);
    A400STATS_ADDMEMBER(result, span_stats, tx_level);
    A400STATS_ADDMEMBER(result, span_stats, rx_level);

    return result;
}

static char *_dahdi_alarm_string(struct dahdi_spaninfo *spaninfo)
{
    static char alarms[50];
    strcpy(alarms, "");
    if (spaninfo->alarms > 0) {
        if (spaninfo->alarms & DAHDI_ALARM_BLUE)
            strcat(alarms,"BLU/");
        if (spaninfo->alarms & DAHDI_ALARM_YELLOW)
            strcat(alarms, "YEL/");
        if (spaninfo->alarms & DAHDI_ALARM_RED)
            strcat(alarms, "RED/");
        if (spaninfo->alarms & DAHDI_ALARM_LOOPBACK)
            strcat(alarms,"LB/");
        if (spaninfo->alarms & DAHDI_ALARM_RECOVER)
            strcat(alarms,"REC/");
        if (spaninfo->alarms & DAHDI_ALARM_NOTOPEN)
            strcat(alarms, "NOP/");
        if (!strlen(alarms))
            strcat(alarms, "UUU/");
        if (strlen(alarms)) {
            /* Strip trailing / */
            alarms[strlen(alarms)-1]='\0';
        }
    } else {
        if (spaninfo->numchans)
            strcpy(alarms, "OK");
        else
            strcpy(alarms, "UNCONFIGURED");
    }
    return alarms;
}

static PyObject *
_dahdi_spaninfo(PyObject *self, int span)
{
    int res;
    struct dahdi_spaninfo spaninfo;
    PyObject *result, *member;

    memset(&spaninfo, '\0', sizeof(spaninfo));
    spaninfo.spanno = span;
    res = ioctl(ctl, DAHDI_SPANSTAT, &spaninfo);
    if( res ) return NULL;

    result = PyDict_New ();
    if (result == NULL)
        return NULL;

    member = Py_BuildValue("i", spaninfo.spanno);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "spanno", member);
    Py_XDECREF (member);

    member = PyString_FromString (spaninfo.name);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "name", member);
    Py_XDECREF (member);

    member = PyString_FromString (spaninfo.desc);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "desc", member);
    Py_XDECREF (member);

    member = PyString_FromString (spaninfo.manufacturer);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "manufacturer", member);
    Py_XDECREF (member);

    member = PyString_FromString (spaninfo.devicetype);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "devicetype", member);
    Py_XDECREF (member);

    member = PyString_FromString (spaninfo.spantype);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "spantype", member);
    Py_XDECREF (member);

    member = Py_BuildValue("i", spaninfo.numchans);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "numchans", member);
    Py_XDECREF (member);

    member = Py_BuildValue("i", spaninfo.numchans);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "numchans", member);
    Py_XDECREF (member);

    member = Py_BuildValue("i", spaninfo.totalchans);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "totalchans", member);
    Py_XDECREF (member);

    member = Py_BuildValue("i", spaninfo.totalspans);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "totalspans", member);
    Py_XDECREF (member);

    member = Py_BuildValue("i", spaninfo.irq);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "irq", member);
    Py_XDECREF (member);

    member = Py_BuildValue("i", spaninfo.alarms);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "alarms", member);
    Py_XDECREF (member);

    member = PyString_FromString (_dahdi_alarm_string(&spaninfo));
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyDict_SetItemString (result, "alarms_string", member);
    Py_XDECREF (member);

    /* TODO: irqmisses */

    /* TODO: cvcount */

    /* TODO: crc4count */

    return result;
}

static PyObject *
dahdi_spaninfo(PyObject *self, PyObject *args)
{
    int span;

    if (!PyArg_ParseTuple(args, "i", &span))
        return NULL;

    return _dahdi_spaninfo(self, span);
}

static PyObject *
dahdi_all_spans(PyObject *self)
{
int i;
PyObject *span, *result;

    result = PyList_New (0);

    for (i=0;i<DAHDI_MAX_SPANS;i++) {
        span = _dahdi_spaninfo(self, i);
        if( span != NULL ) {
            PyList_Append (result, span);
        }
    }

    return result;
}

static PyObject *
dahdi_spanalarms(PyObject *self, PyObject *args)
{
    int span, res;
    struct dahdi_spaninfo spaninfo;
    PyObject *result, *member;

    memset(&spaninfo, '\0', sizeof(spaninfo));

    if (!PyArg_ParseTuple(args, "i", &span))
        return NULL;

    spaninfo.spanno = span;
    res = ioctl(ctl, DAHDI_SPANSTAT, &spaninfo);
    if( res ) return NULL;

    result = PyList_New (0);
    if (result == NULL)
        return NULL;

    member = Py_BuildValue("i", spaninfo.alarms);
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyList_Append (result, member);
    Py_XDECREF (member);

    member = Py_BuildValue("s", _dahdi_alarm_string(&spaninfo));
    if( member == NULL ) {
        Py_XDECREF (result);
        return NULL;
    }
    PyList_Append (result, member);
    Py_XDECREF (member);

    return result;
}

static PyObject *
dahdi_close(PyObject *self)
{
	close(ctl);

	return Py_None;
}

static PyMethodDef AP400Methods[] = {
    {"ap400_alarms",  ap400_alarms, METH_VARARGS, "Get AP400 alarms"},
    {"find_first_channel",  ap400_find_first_channel, METH_VARARGS, "Get the first channel number of a DAHDI span"},
    {"hwinfo",  ap400_hwinfo, METH_VARARGS, "Get AP400 hardware information"},
    {"stats",  ap400_spats, METH_VARARGS, "Get AP400 span statistics"},
    {"spaninfo",  dahdi_spaninfo, METH_VARARGS, "Get DAHDI spaninfo members as a dict for a single span"},
    {"allspans",  (PyCFunction) dahdi_all_spans, METH_NOARGS, "Retuns a list of spaninfo dictionaries for all spans found on the system"},
    {"dahdi_alarms",  dahdi_spanalarms, METH_VARARGS, "Get DAHDI active alarms as a list [VALUE, STRING]"},
    {"close",  (PyCFunction) dahdi_close, METH_NOARGS, "Closes DAHDI file descriptor"},
    {NULL, NULL, 0, NULL}        /* Sentinel */
};

PyMODINIT_FUNC
initap400(void)
{
	(void) Py_InitModule("ap400", AP400Methods);

    ctl = open("/dev/dahdi/ctl", O_RDWR);
}

int main(int argc, char *argv[])
{
    /* Pass argv[0] to the Python interpreter */
    Py_SetProgramName(argv[0]);

    /* Initialize the Python interpreter.  Required. */
    Py_Initialize();

    /* Add a static module */
    initap400();

    return 0;
}
