/*
 * configuration.c
 *
 *  Created on: May 23, 2011
 *      Author: Rodrigo Ferreira <rferreira@pd3.com.br>
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>

#include "configuration.h"

#define INITIAL_HTABLE_SIZE 	13
#define MAX_LINE_LENGTH	1024

struct config *router_factory_conf = NULL;
struct config *router_global_param = NULL;

static void _destroy_config_data(void *data)
{
	char *conf = data;
	free(conf);
}

char *config_get_value(struct config *conf, const char *key)
{
	char *value = NULL;

	if (conf == NULL)
		return NULL;

	if (conf->initialized == 0)
		return NULL;

	value = (char *) htable_lookup(conf->config_table, (char *) key);
	return value;
}

void config_destroy(struct config **conf)
{
	if (*conf == NULL)
		return;

	htable_destroy((*conf)->config_table);
	free(*conf);
	*conf = NULL;
}

void config_init(struct config **conf)
{
	struct config *c;

	if ((c = malloc(sizeof(struct config))) == NULL) {
		err(1, "malloc");
	}

	c->config_table = htable_create(INITIAL_HTABLE_SIZE, _destroy_config_data);
	c->errors = 0;
	c->initialized = 0;

	*conf = c;
	return;
}

static void _remove_backslash(char *buf)
{
	int i;
	char* pos;

	pos = strchr(buf, '\\');
	while (pos != NULL) {
		i = 0;
		while (pos[i] != 0) {
			pos[i] = pos[i + 1];
			i++;
		}
		if (pos[0] == '\\')
			pos = strchr(pos + 1, '\\');
		else
			pos = strchr(pos, '\\');
	}
}

static int _check_key(char *key)
{
	char *temp;
	int i;

	temp = key + 1;
	while (temp[0] != 0) {
		i = 0;
		while (temp[i] != ' ' && temp[i] != '\t') {
			if (temp[i] == 0)
				return 0;
			i++;
		}
		if (temp[i - 1] == '\\')
			temp = temp + i + 1;
		else
			return 1;
	}
	return 0;
}

int config_read_conf(struct config *conf, const char *file)
{
	FILE *fp;
	int read_bytes, i, line_counter, error, multiline;
	size_t length;
	char *line, *left, *right, *eq, *key, *value;
	char buf[MAX_LINE_LENGTH];

	if (conf == NULL)
		return -1;

	if (conf->initialized)
		return 0;

	if ((fp = fopen(file, "r")) == NULL) {
		return -2;
	}

	error = 0;
	line_counter = 0;
	multiline = 0;
	length = MAX_LINE_LENGTH;

	if ((line = malloc(length * sizeof(char))) == NULL) {
		err(1, "malloc");
	}

	while ((read_bytes = getline(&line, &length, fp)) != -1) {
		line_counter++;

		if (read_bytes < 1)
			continue;

		i = 0;
		while (line[i] == ' ' || line[i] == '\t')
			i++;

		if (line[i] == '#' || line[i] == '\n')
			continue;

		left = line + i;
		if (multiline)
			strcat(buf, left);
		else
			strcpy(buf, left);

		eq = strchr(buf, ':');
		if (eq == NULL) {
			error++;
			continue;
		}

		i = 1;
		while (!(eq[i] == '\n' || eq[i] == '\0')) {
			if (eq[i] == '#' && eq[i - 1] != '\\')
				break;
			i++;
		}
		eq[i] = '\0';

		right = &eq[i];
		i = right - buf;
		if (buf[i - 1] == '\\') {
			buf[i] = '\0';
			multiline = 1;
			continue;
		} else {
			while (buf[i - 1] == ' ' || buf[i - 1] == '\t')
				i--;
			buf[i] = '\0';
			multiline = 0;
		}

		eq = strchr(buf, ':');
		i = eq - buf;
		while (buf[i - 1] == ' ' || buf[i - 1] == '\t')
			i--;
		key = strndup(buf, i);
		if (_check_key(key)) {
			error++;
			continue;
		}
		_remove_backslash(key);

		i = 1;
		while (eq[i] == ' ' || eq[i] == '\t')
			i++;

		left = eq + i;
		_remove_backslash(left);
		value = strdup(left);

		htable_insert(conf->config_table, key, value);
		free(key);
	}

	fclose(fp);
	free(line);

	conf->errors = error;
	conf->initialized = 1;

	return (-conf->errors);
}

int config_get_errors(struct config *conf)
{
	return conf->errors;
}

int config_save_conf(struct config *conf, const char* filepath)
{
	FILE *fp;
	struct _htable_node *hnode;
	int i;
	htable *ht = conf->config_table;

	if ((fp = fopen(filepath, "w")) == NULL)
		return -1;

	for (i = 0; i < ht->size; i++) {
		hnode = ht->table[i];
		while (hnode != NULL) {
			fprintf(fp, "%s = %s \n", hnode->key, (char*) hnode->data);
			hnode = hnode->next;
		}
	}
	fclose(fp);

	return 0;
}

int config_read_defaults(struct config *conf, const char *file)
{
	config_read_conf(conf, file);
	conf->initialized = 0;
	return (-conf->errors);
}

int configuration_update_keys(struct config *old_conf,
                              char *old_version_match,
                              const char* filepath,
                              char *new_version_match,
                              char *version_key,
                              char **replace_array,
                              int replace_count)
{
	struct config *new_conf = NULL;
	char *old_version, *new_version;
	char *old_key, *new_key, *old_value;
	int i, err_count = 0, old_match = 0, new_match = 0;

	/* check all parameters */
	if ((old_conf == NULL) || !old_conf->initialized)
		return -1;
	if (filepath == NULL)
		return -1;
	if ((replace_array == NULL) || (replace_count <= 0))
		return -1;

	/* check old version */
	old_version = config_get_value(old_conf, version_key);
	if (old_version == NULL) {
		if (old_version_match == NULL)
			old_match = 1;
	} else if (old_version_match != NULL) {
		if (strcmp(old_version, old_version_match) == 0)
			old_match = 1;
	}
	if (!old_match) {
		return 0;
	}

	/* read new configuration */
	config_init(&new_conf);
	err_count = config_read_conf(new_conf, filepath);
	if (err_count < 0) {
		config_destroy(&new_conf);
		return err_count;
	}

	/* check new version */
	new_version = config_get_value(new_conf, version_key);
	if (new_version == NULL) {
		if (new_version_match == NULL)
			new_match = 1;
	} else if (new_version_match != NULL) {
		if (strcmp(new_version, new_version_match) == 0)
			new_match = 1;
	}
	if (!new_match) {
		config_destroy(&new_conf);
		return 0;
	}

	/* Adds the values of the legacy keys with new keys  */
	for (i = 0; i < replace_count; i++) {
		old_key = replace_array[2 * i];
		new_key = replace_array[2 * i + 1];

		old_value = config_get_value(old_conf, old_key);
		htable_insert(old_conf->config_table, new_key,
		                strdup(old_value));
	}

	config_destroy(&new_conf);
	return replace_count;
}
